@@ -23,21 +23,6 @@ type Workflow struct {
 	Defaults Defaults          `yaml:"defaults"`
 }
 
-// CompositeRestrictions is the structure to control what is allowed in composite actions
-type CompositeRestrictions struct {
-	AllowCompositeUses            bool
-	AllowCompositeIf              bool
-	AllowCompositeContinueOnError bool
-}
-
-func defaultCompositeRestrictions() *CompositeRestrictions {
-	return &CompositeRestrictions{
-		AllowCompositeUses:            true,
-		AllowCompositeIf:              true,
-		AllowCompositeContinueOnError: false,
-	}
-}
-
 // On events for the workflow
 func (w *Workflow) On() []string {
 	switch w.RawOn.Kind {
@@ -84,6 +69,7 @@ type Job struct {
 	RawContainer   yaml.Node                 `yaml:"container"`
 	Defaults       Defaults                  `yaml:"defaults"`
 	Outputs        map[string]string         `yaml:"outputs"`
+	Uses           string                    `yaml:"uses"`
 	Result         string
 }
 
@@ -238,26 +224,37 @@ func (j *Job) GetMatrixes() []map[string]interface{} {
 
 		if m := j.Matrix(); m != nil {
 			includes := make([]map[string]interface{}, 0)
+			extraIncludes := make([]map[string]interface{}, 0)
 			for _, v := range m["include"] {
 				switch t := v.(type) {
 				case []interface{}:
 					for _, i := range t {
 						i := i.(map[string]interface{})
+						extraInclude := true
 						for k := range i {
 							if _, ok := m[k]; ok {
 								includes = append(includes, i)
+								extraInclude = false
 								break
 							}
 						}
+						if extraInclude {
+							extraIncludes = append(extraIncludes, i)
+						}
 					}
 				case interface{}:
 					v := v.(map[string]interface{})
+					extraInclude := true
 					for k := range v {
 						if _, ok := m[k]; ok {
 							includes = append(includes, v)
+							extraInclude = false
 							break
 						}
 					}
+					if extraInclude {
+						extraIncludes = append(extraIncludes, v)
+					}
 				}
 			}
 			delete(m, "include")
@@ -288,9 +285,27 @@ func (j *Job) GetMatrixes() []map[string]interface{} {
 				matrixes = append(matrixes, matrix)
 			}
 			for _, include := range includes {
+				matched := false
+				for _, matrix := range matrixes {
+					if commonKeysMatch2(matrix, include, m) {
+						matched = true
+						log.Debugf("Adding include values '%v' to existing entry", include)
+						for k, v := range include {
+							matrix[k] = v
+						}
+					}
+				}
+				if !matched {
+					extraIncludes = append(extraIncludes, include)
+				}
+			}
+			for _, include := range extraIncludes {
 				log.Debugf("Adding include '%v'", include)
 				matrixes = append(matrixes, include)
 			}
+			if len(matrixes) == 0 {
+				matrixes = append(matrixes, make(map[string]interface{}))
+			}
 		} else {
 			matrixes = append(matrixes, make(map[string]interface{}))
 		}
@@ -309,6 +324,16 @@ func commonKeysMatch(a map[string]interface{}, b map[string]interface{}) bool {
 	return true
 }
 
+func commonKeysMatch2(a map[string]interface{}, b map[string]interface{}, m map[string][]interface{}) bool {
+	for aKey, aVal := range a {
+		_, useKey := m[aKey]
+		if bVal, ok := b[aKey]; useKey && ok && !reflect.DeepEqual(aVal, bVal) {
+			return false
+		}
+	}
+	return true
+}
+
 // ContainerSpec is the specification of the container to use for the job
 type ContainerSpec struct {
 	Image       string            `yaml:"image"`
@@ -335,7 +360,7 @@ type Step struct {
 	Env              yaml.Node         `yaml:"env"`
 	With             map[string]string `yaml:"with"`
 	ContinueOnError  bool              `yaml:"continue-on-error"`
-	TimeoutMinutes   int64             `yaml:"timeout-minutes"`
+	TimeoutMinutes   string            `yaml:"timeout-minutes"`
 }
 
 // String gets the name of step
@@ -413,6 +438,10 @@ const (
 
 // Type returns the type of the step
 func (s *Step) Type() StepType {
+	if s.Run == "" && s.Uses == "" {
+		return StepTypeInvalid
+	}
+
 	if s.Run != "" {
 		if s.Uses != "" {
 			return StepTypeInvalid
@@ -426,22 +455,6 @@ func (s *Step) Type() StepType {
 	return StepTypeUsesActionRemote
 }
 
-func (s *Step) Validate(config *CompositeRestrictions) error {
-	if config == nil {
-		config = defaultCompositeRestrictions()
-	}
-	if s.Type() != StepTypeRun && !config.AllowCompositeUses {
-		return fmt.Errorf("(StepID: %s): Unexpected value 'uses'", s.String())
-	} else if s.Type() == StepTypeRun && s.Shell == "" {
-		return fmt.Errorf("(StepID: %s): Required property is missing: 'shell'", s.String())
-	} else if !s.If.IsZero() && !config.AllowCompositeIf {
-		return fmt.Errorf("(StepID: %s): Property is not available: 'if'", s.String())
-	} else if s.ContinueOnError && !config.AllowCompositeContinueOnError {
-		return fmt.Errorf("(StepID: %s): Property is not available: 'continue-on-error'", s.String())
-	}
-	return nil
-}
-
 // ReadWorkflow returns a list of jobs for a given workflow file reader
 func ReadWorkflow(in io.Reader) (*Workflow, error) {
 	w := new(Workflow)
