@@ -10,16 +10,16 @@ import (
 	"runtime"
 	"strings"
 
-	"github.com/google/shlex"
+	"github.com/kballard/go-shellquote"
+	"github.com/mitchellh/go-homedir"
+	"github.com/opencontainers/selinux/go-selinux"
+	log "github.com/sirupsen/logrus"
 	"github.com/spf13/pflag"
 
-	"github.com/mitchellh/go-homedir"
-	log "github.com/sirupsen/logrus"
-
-	selinux "github.com/opencontainers/selinux/go-selinux"
-
 	"github.com/nektos/act/pkg/common"
+	"github.com/nektos/act/pkg/common/git"
 	"github.com/nektos/act/pkg/container"
+	"github.com/nektos/act/pkg/exprparser"
 	"github.com/nektos/act/pkg/model"
 )
 
@@ -43,7 +43,6 @@ type RunContext struct {
 	ActionPath       string
 	ActionRef        string
 	ActionRepository string
-	Composite        *model.Action
 	Inputs           map[string]interface{}
 	Parent           *RunContext
 	Masks            []string
@@ -53,16 +52,6 @@ func (rc *RunContext) AddMask(mask string) {
 	rc.Masks = append(rc.Masks, mask)
 }
 
-func (rc *RunContext) Clone() *RunContext {
-	clone := *rc
-	clone.CurrentStep = ""
-	clone.Composite = nil
-	clone.Inputs = nil
-	clone.StepResults = make(map[string]*model.StepResult)
-	clone.Parent = rc
-	return &clone
-}
-
 type MappableOutput struct {
 	StepID     string
 	OutputName string
@@ -75,7 +64,7 @@ func (rc *RunContext) String() string {
 // GetEnv returns the env for the context
 func (rc *RunContext) GetEnv() map[string]string {
 	if rc.Env == nil {
-		rc.Env = mergeMaps(rc.Config.Env, rc.Run.Workflow.Env, rc.Run.Job().Environment())
+		rc.Env = mergeMaps(rc.Run.Workflow.Env, rc.Run.Job().Environment(), rc.Config.Env)
 	}
 	rc.Env["ACT"] = "true"
 	return rc.Env
@@ -102,6 +91,21 @@ func (rc *RunContext) GetBindsAndMounts() ([]string, map[string]string) {
 		name + "-env":   ActPath,
 	}
 
+	if job := rc.Run.Job(); job != nil {
+		if container := job.Container(); container != nil {
+			for _, v := range container.Volumes {
+				if !strings.Contains(v, ":") || filepath.IsAbs(v) {
+					// Bind anonymous volume or host file.
+					binds = append(binds, v)
+				} else {
+					// Mount existing volume.
+					paths := strings.SplitN(v, ":", 2)
+					mounts[paths[0]] = paths[1]
+				}
+			}
+		}
+	}
+
 	if rc.Config.BindWorkdir {
 		bindModifiers := ""
 		if runtime.GOOS == "darwin" {
@@ -119,11 +123,11 @@ func (rc *RunContext) GetBindsAndMounts() ([]string, map[string]string) {
 }
 
 func (rc *RunContext) startJobContainer() common.Executor {
-	image := rc.platformImage()
-	hostname := rc.hostname()
-
 	return func(ctx context.Context) error {
-		rawLogger := common.Logger(ctx).WithField("raw_output", true)
+		logger := common.Logger(ctx)
+		image := rc.platformImage(ctx)
+		hostname := rc.hostname(ctx)
+		rawLogger := logger.WithField("raw_output", true)
 		logWriter := common.NewLineWriter(rc.commandHandler(ctx), func(s string) bool {
 			if rc.Config.LogOutput {
 				rawLogger.Infof("%s", s)
@@ -133,12 +137,12 @@ func (rc *RunContext) startJobContainer() common.Executor {
 			return true
 		})
 
-		username, password, err := rc.handleCredentials()
+		username, password, err := rc.handleCredentials(ctx)
 		if err != nil {
 			return fmt.Errorf("failed to handle credentials: %s", err)
 		}
 
-		common.Logger(ctx).Infof("\U0001f680  Start image=%s", image)
+		logger.Infof("\U0001f680  Start image=%s", image)
 		name := rc.jobContainerName()
 
 		envList := make([]string, 0)
@@ -169,13 +173,6 @@ func (rc *RunContext) startJobContainer() common.Executor {
 			Hostname:    hostname,
 		})
 
-		var copyWorkspace bool
-		var copyToPath string
-		if !rc.Config.BindWorkdir {
-			copyToPath, copyWorkspace = rc.localCheckoutPath()
-			copyToPath = filepath.Join(rc.Config.ContainerWorkdir(), copyToPath)
-		}
-
 		return common.NewPipelineExecutor(
 			rc.JobContainer.Pull(rc.Config.ForcePull),
 			rc.stopJobContainer(),
@@ -183,8 +180,6 @@ func (rc *RunContext) startJobContainer() common.Executor {
 			rc.JobContainer.Start(false),
 			rc.JobContainer.UpdateFromImageEnv(&rc.Env),
 			rc.JobContainer.UpdateFromEnv("/etc/environment", &rc.Env),
-			rc.JobContainer.Exec([]string{"mkdir", "-m", "0777", "-p", ActPath}, rc.Env, "root", ""),
-			rc.JobContainer.CopyDir(copyToPath, rc.Config.Workdir+string(filepath.Separator)+".", rc.Config.UseGitIgnore).IfBool(copyWorkspace),
 			rc.JobContainer.Copy(ActPath+"/", &container.FileEntry{
 				Name: "workflow/event.json",
 				Mode: 0644,
@@ -213,7 +208,8 @@ func (rc *RunContext) stopJobContainer() common.Executor {
 	return func(ctx context.Context) error {
 		if rc.JobContainer != nil && !rc.Config.ReuseContainers {
 			return rc.JobContainer.Remove().
-				Then(container.NewDockerVolumeRemoveExecutor(rc.jobContainerName(), false))(ctx)
+				Then(container.NewDockerVolumeRemoveExecutor(rc.jobContainerName(), false)).
+				Then(container.NewDockerVolumeRemoveExecutor(rc.jobContainerName()+"-env", false))(ctx)
 		}
 		return nil
 	}
@@ -238,9 +234,9 @@ func (rc *RunContext) ActionCacheDir() string {
 // Interpolate outputs after a job is done
 func (rc *RunContext) interpolateOutputs() common.Executor {
 	return func(ctx context.Context) error {
-		ee := rc.NewExpressionEvaluator()
+		ee := rc.NewExpressionEvaluator(ctx)
 		for k, v := range rc.Run.Job().Outputs {
-			interpolated := ee.Interpolate(v)
+			interpolated := ee.Interpolate(ctx, v)
 			if v != interpolated {
 				rc.Run.Job().Outputs[k] = interpolated
 			}
@@ -287,109 +283,27 @@ func (rc *RunContext) Executor() common.Executor {
 		}
 
 		if isEnabled {
-			return newJobExecutor(rc)(ctx)
+			return newJobExecutor(rc, &stepFactoryImpl{}, rc)(ctx)
 		}
 
 		return nil
 	}
 }
 
-// Executor returns a pipeline executor for all the steps in the job
-func (rc *RunContext) CompositeExecutor() common.Executor {
-	steps := make([]common.Executor, 0)
-
-	for i, step := range rc.Composite.Runs.Steps {
-		if step.ID == "" {
-			step.ID = fmt.Sprintf("%d", i)
-		}
-		stepcopy := step
-		stepExec := rc.newStepExecutor(&stepcopy)
-		steps = append(steps, func(ctx context.Context) error {
-			err := stepExec(ctx)
-			if err != nil {
-				common.Logger(ctx).Errorf("%v", err)
-				common.SetJobError(ctx, err)
-			} else if ctx.Err() != nil {
-				common.Logger(ctx).Errorf("%v", ctx.Err())
-				common.SetJobError(ctx, ctx.Err())
-			}
-			return nil
-		})
-	}
-
-	steps = append(steps, common.JobError)
-	return func(ctx context.Context) error {
-		return common.NewPipelineExecutor(steps...)(common.WithJobErrorContainer(ctx))
-	}
-}
-
-func (rc *RunContext) newStepExecutor(step *model.Step) common.Executor {
-	sc := &StepContext{
-		RunContext: rc,
-		Step:       step,
-	}
-	return func(ctx context.Context) error {
-		rc.CurrentStep = sc.Step.ID
-		rc.StepResults[rc.CurrentStep] = &model.StepResult{
-			Outcome:    model.StepStatusSuccess,
-			Conclusion: model.StepStatusSuccess,
-			Outputs:    make(map[string]string),
-		}
-
-		runStep, err := sc.isEnabled(ctx)
-		if err != nil {
-			rc.StepResults[rc.CurrentStep].Conclusion = model.StepStatusFailure
-			rc.StepResults[rc.CurrentStep].Outcome = model.StepStatusFailure
-			return err
-		}
-
-		if !runStep {
-			log.Debugf("Skipping step '%s' due to '%s'", sc.Step.String(), sc.Step.If.Value)
-			rc.StepResults[rc.CurrentStep].Conclusion = model.StepStatusSkipped
-			rc.StepResults[rc.CurrentStep].Outcome = model.StepStatusSkipped
-			return nil
-		}
-
-		exprEval, err := sc.setupEnv(ctx)
-		if err != nil {
-			return err
-		}
-		rc.ExprEval = exprEval
-
-		common.Logger(ctx).Infof("\u2B50  Run %s", sc.Step)
-		err = sc.Executor(ctx)(ctx)
-		if err == nil {
-			common.Logger(ctx).Infof("  \u2705  Success - %s", sc.Step)
-		} else {
-			common.Logger(ctx).Errorf("  \u274C  Failure - %s", sc.Step)
-
-			rc.StepResults[rc.CurrentStep].Outcome = model.StepStatusFailure
-			if sc.Step.ContinueOnError {
-				common.Logger(ctx).Infof("Failed but continue next step")
-				err = nil
-				rc.StepResults[rc.CurrentStep].Conclusion = model.StepStatusSuccess
-			} else {
-				rc.StepResults[rc.CurrentStep].Conclusion = model.StepStatusFailure
-			}
-		}
-		return err
-	}
-}
-
-func (rc *RunContext) platformImage() string {
+func (rc *RunContext) platformImage(ctx context.Context) string {
 	job := rc.Run.Job()
 
 	c := job.Container()
 	if c != nil {
-		return rc.ExprEval.Interpolate(c.Image)
+		return rc.ExprEval.Interpolate(ctx, c.Image)
 	}
 
 	if job.RunsOn() == nil {
-		log.Errorf("'runs-on' key not defined in %s", rc.String())
+		common.Logger(ctx).Errorf("'runs-on' key not defined in %s", rc.String())
 	}
 
 	for _, runnerLabel := range job.RunsOn() {
-		platformName := rc.ExprEval.Interpolate(runnerLabel)
+		platformName := rc.ExprEval.Interpolate(ctx, runnerLabel)
 		image := rc.Config.Platforms[strings.ToLower(platformName)]
 		if image != "" {
 			return image
@@ -399,7 +313,8 @@ func (rc *RunContext) platformImage() string {
 	return ""
 }
 
-func (rc *RunContext) hostname() string {
+func (rc *RunContext) hostname(ctx context.Context) string {
+	logger := common.Logger(ctx)
 	job := rc.Run.Job()
 	c := job.Container()
 	if c == nil {
@@ -408,14 +323,14 @@ func (rc *RunContext) hostname() string {
 
 	optionsFlags := pflag.NewFlagSet("container_options", pflag.ContinueOnError)
 	hostname := optionsFlags.StringP("hostname", "h", "", "")
-	optionsArgs, err := shlex.Split(c.Options)
+	optionsArgs, err := shellquote.Split(c.Options)
 	if err != nil {
-		log.Warnf("Cannot parse container options: %s", c.Options)
+		logger.Warnf("Cannot parse container options: %s", c.Options)
 		return ""
 	}
 	err = optionsFlags.Parse(optionsArgs)
 	if err != nil {
-		log.Warnf("Cannot parse container options: %s", c.Options)
+		logger.Warnf("Cannot parse container options: %s", c.Options)
 		return ""
 	}
 	return *hostname
@@ -424,23 +339,23 @@ func (rc *RunContext) hostname() string {
 func (rc *RunContext) isEnabled(ctx context.Context) (bool, error) {
 	job := rc.Run.Job()
 	l := common.Logger(ctx)
-	runJob, err := EvalBool(rc.ExprEval, job.If.Value)
+	runJob, err := EvalBool(ctx, rc.ExprEval, job.If.Value, exprparser.DefaultStatusCheckSuccess)
 	if err != nil {
 		return false, fmt.Errorf("  \u274C  Error in if-expression: \"if: %s\" (%s)", job.If.Value, err)
 	}
 	if !runJob {
-		l.Debugf("Skipping job '%s' due to '%s'", job.Name, job.If.Value)
+		l.WithField("jobResult", "skipped").Debugf("Skipping job '%s' due to '%s'", job.Name, job.If.Value)
 		return false, nil
 	}
 
-	img := rc.platformImage()
+	img := rc.platformImage(ctx)
 	if img == "" {
 		if job.RunsOn() == nil {
-			log.Errorf("'runs-on' key not defined in %s", rc.String())
+			l.Errorf("'runs-on' key not defined in %s", rc.String())
 		}
 
 		for _, runnerLabel := range job.RunsOn() {
-			platformName := rc.ExprEval.Interpolate(runnerLabel)
+			platformName := rc.ExprEval.Interpolate(ctx, runnerLabel)
 			l.Infof("\U0001F6A7  Skipping unsupported platform -- Try running with `-P %+v=...`", platformName)
 		}
 		return false, nil
@@ -508,7 +423,8 @@ func (rc *RunContext) getStepsContext() map[string]*model.StepResult {
 	return rc.StepResults
 }
 
-func (rc *RunContext) getGithubContext() *model.GithubContext {
+func (rc *RunContext) getGithubContext(ctx context.Context) *model.GithubContext {
+	logger := common.Logger(ctx)
 	ghc := &model.GithubContext{
 		Event:            make(map[string]interface{}),
 		EventPath:        ActPath + "/workflow/event.json",
@@ -519,7 +435,7 @@ func (rc *RunContext) getGithubContext() *model.GithubContext {
 		EventName:        rc.Config.EventName,
 		Workspace:        rc.Config.ContainerWorkdir(),
 		Action:           rc.CurrentStep,
-		Token:            rc.Config.Secrets["GITHUB_TOKEN"],
+		Token:            rc.Config.Token,
 		ActionPath:       rc.ActionPath,
 		ActionRef:        rc.ActionRef,
 		ActionRepository: rc.ActionRepository,
@@ -552,9 +468,9 @@ func (rc *RunContext) getGithubContext() *model.GithubContext {
 	}
 
 	repoPath := rc.Config.Workdir
-	repo, err := common.FindGithubRepo(repoPath, rc.Config.GitHubInstance)
+	repo, err := git.FindGithubRepo(ctx, repoPath, rc.Config.GitHubInstance, rc.Config.RemoteName)
 	if err != nil {
-		log.Warningf("unable to get git repo: %v", err)
+		logger.Warningf("unable to get git repo: %v", err)
 	} else {
 		ghc.Repository = repo
 		if ghc.RepositoryOwner == "" {
@@ -565,7 +481,7 @@ func (rc *RunContext) getGithubContext() *model.GithubContext {
 	if rc.EventJSON != "" {
 		err = json.Unmarshal([]byte(rc.EventJSON), &ghc.Event)
 		if err != nil {
-			log.Errorf("Unable to Unmarshal event '%s': %v", rc.EventJSON, err)
+			logger.Errorf("Unable to Unmarshal event '%s': %v", rc.EventJSON, err)
 		}
 	}
 
@@ -574,7 +490,16 @@ func (rc *RunContext) getGithubContext() *model.GithubContext {
 		ghc.HeadRef = asString(nestedMapLookup(ghc.Event, "pull_request", "head", "ref"))
 	}
 
-	ghc.SetRefAndSha(rc.Config.DefaultBranch, repoPath)
+	ghc.SetRefAndSha(ctx, rc.Config.DefaultBranch, repoPath)
+
+	// https://docs.github.com/en/actions/learn-github-actions/environment-variables
+	if strings.HasPrefix(ghc.Ref, "refs/tags/") {
+		ghc.RefType = "tag"
+		ghc.RefName = ghc.Ref[len("refs/tags/"):]
+	} else if strings.HasPrefix(ghc.Ref, "refs/heads/") {
+		ghc.RefType = "branch"
+		ghc.RefName = ghc.Ref[len("refs/heads/"):]
+	}
 
 	return ghc
 }
@@ -631,8 +556,8 @@ func nestedMapLookup(m map[string]interface{}, ks ...string) (rval interface{})
 	}
 }
 
-func (rc *RunContext) withGithubEnv(env map[string]string) map[string]string {
-	github := rc.getGithubContext()
+func (rc *RunContext) withGithubEnv(ctx context.Context, env map[string]string) map[string]string {
+	github := rc.getGithubContext(ctx)
 	env["CI"] = "true"
 	env["GITHUB_ENV"] = ActPath + "/workflow/envs.txt"
 	env["GITHUB_PATH"] = ActPath + "/workflow/paths.txt"
@@ -651,12 +576,12 @@ func (rc *RunContext) withGithubEnv(env map[string]string) map[string]string {
 	env["GITHUB_WORKSPACE"] = github.Workspace
 	env["GITHUB_SHA"] = github.Sha
 	env["GITHUB_REF"] = github.Ref
+	env["GITHUB_REF_NAME"] = github.RefName
+	env["GITHUB_REF_TYPE"] = github.RefType
 	env["GITHUB_TOKEN"] = github.Token
 	env["GITHUB_SERVER_URL"] = "https://github.com"
 	env["GITHUB_API_URL"] = "https://api.github.com"
 	env["GITHUB_GRAPHQL_URL"] = "https://api.github.com/graphql"
-	env["GITHUB_ACTION_REF"] = github.ActionRef
-	env["GITHUB_ACTION_REPOSITORY"] = github.ActionRepository
 	env["GITHUB_BASE_REF"] = github.BaseRef
 	env["GITHUB_HEAD_REF"] = github.HeadRef
 	env["GITHUB_JOB"] = rc.JobName
@@ -677,7 +602,7 @@ func (rc *RunContext) withGithubEnv(env map[string]string) map[string]string {
 	job := rc.Run.Job()
 	if job.RunsOn() != nil {
 		for _, runnerLabel := range job.RunsOn() {
-			platformName := rc.ExprEval.Interpolate(runnerLabel)
+			platformName := rc.ExprEval.Interpolate(ctx, runnerLabel)
 			if platformName != "" {
 				if platformName == "ubuntu-latest" {
 					// hardcode current ubuntu-latest since we have no way to check that 'on the fly'
@@ -707,21 +632,7 @@ func setActionRuntimeVars(rc *RunContext, env map[string]string) {
 	env["ACTIONS_RUNTIME_TOKEN"] = actionsRuntimeToken
 }
 
-func (rc *RunContext) localCheckoutPath() (string, bool) {
-	if rc.Config.NoSkipCheckout {
-		return "", false
-	}
-
-	ghContext := rc.getGithubContext()
-	for _, step := range rc.Run.Job().Steps {
-		if isLocalCheckout(ghContext, step) {
-			return step.With["path"], true
-		}
-	}
-	return "", false
-}
-
-func (rc *RunContext) handleCredentials() (username, password string, err error) {
+func (rc *RunContext) handleCredentials(ctx context.Context) (username, password string, err error) {
 	// TODO: remove below 2 lines when we can release act with breaking changes
 	username = rc.Config.Secrets["DOCKER_USERNAME"]
 	password = rc.Config.Secrets["DOCKER_PASSWORD"]
@@ -736,12 +647,12 @@ func (rc *RunContext) handleCredentials() (username, password string, err error)
 		return
 	}
 
-	ee := rc.NewExpressionEvaluator()
-	if username = ee.Interpolate(container.Credentials["username"]); username == "" {
+	ee := rc.NewExpressionEvaluator(ctx)
+	if username = ee.Interpolate(ctx, container.Credentials["username"]); username == "" {
 		err = fmt.Errorf("failed to interpolate container.credentials.username")
 		return
 	}
-	if password = ee.Interpolate(container.Credentials["password"]); password == "" {
+	if password = ee.Interpolate(ctx, container.Credentials["password"]); password == "" {
 		err = fmt.Errorf("failed to interpolate container.credentials.password")
 		return
 	}
